self.__BUILD_MANIFEST=function(s,c){return{__rewrites:{afterFiles:[],beforeFiles:[],fallback:[]},"/":[s,"static/chunks/639-7b872ad837b4b411.js",c,"static/css/3bebd5c08d1329c4.css","static/chunks/pages/index-68765863946878e7.js"],"/_error":["static/chunks/pages/_error-7a92967bea80186d.js"],"/benchmarks":[s,c,"static/css/ae560ac573ba7271.css","static/chunks/pages/benchmarks-10fafa2e21e1dbea.js"],sortedPages:["/","/_app","/_error","/benchmarks"]}}("static/chunks/3-16995fae4ad08ab9.js","static/chunks/736-88d005ed18d0400b.js"),self.__BUILD_MANIFEST_CB&&self.__BUILD_MANIFEST_CB();